#include "internal_shared.hpp"

// #include <spdlog/async.h>
// #include <spdlog/cfg/env.h>
// #include <spdlog/logger.h>
// #include <spdlog/sinks/basic_file_sink.h>
// #include <spdlog/sinks/rotating_file_sink.h>
// #include <spdlog/sinks/stdout_color_sinks.h>
// #include <spdlog/spdlog.h>

#include <memory>
#include <mutex>
#include <sole.hpp>

namespace tsdb_shared {

std::once_flag tsdb_env_init_flag;

std::shared_ptr<mylogger> g_logger;

static mylogger::loglevel load_env_level() {
    auto level_ptr = std::getenv("TSDBLOG_LEVEL");
    if (!level_ptr) {
        return mylogger::loglevel::info;
    }
    string_view level(level_ptr);
    if (auto pos = level.find_first_of(",;"); pos != string_view::npos) {
        level = level.substr(0, pos);
    }
    if (level == "trace") {
        return mylogger::loglevel::trace;
    } else if (level == "debug") {
        return mylogger::loglevel::debug;
    } else if (level == "info") {
        return mylogger::loglevel::info;
    } else if (level == "warn") {
        return mylogger::loglevel::warn;
    } else if (level == "error") {
        return mylogger::loglevel::error;
    } else if (level == "critical") {
        return mylogger::loglevel::critical;
    } else {
        return mylogger::loglevel::info;
    }
}

void init_tsdb_env_inner() {
    // Initialize the TSDB environment

    auto env_is_set = [](const char* name) {
        auto value = std::getenv(name);
        return value && value[0] != '\0';
    };
    auto env_is_true = [](const char* name) {
        auto value = std::getenv(name);
        return value && value[0] != '\0' && value[0] != '0';
    };

    auto log_level        = load_env_level();
    auto use_file_logging = env_is_set("TSDB_USE_FILE_LOGGING") ? env_is_true("TSDB_USE_FILE_LOGGING") : true;
    auto log_path         = std::getenv("TSDB_LOG_PATH") ?: "tsdb.log";
    std::shared_ptr<mylogger> logger;
    if (use_file_logging) {
        // logger = spdlog::basic_logger_mt<spdlog::async_factory>("tsdb", log_path);
        logger =
            std::make_shared<mylogger>("tsdb", log_level, bin_file(log_path, "ab"), bin_file(std::string(log_path) + ".error.log", "ab"));
    } else {
        // logger = spdlog::stdout_color_mt<spdlog::async_factory>("tsdb");
        logger = std::make_shared<mylogger>("tsdb", log_level);
    }
    g_logger = std::move(logger);

    g_logger->warn("TSDB logger initialized, compile datetime: {}, cpp version: {}", __DATE__ " " __TIME__, __cplusplus);
}

mylogger& logger() {
    return *g_logger;
}

void init_tsdb_env() {
    std::call_once(tsdb_env_init_flag, init_tsdb_env_inner);
}

std::string random_uuid() {
    return sole::uuid4().str();
}

}  // namespace tsdb_shared
