// Don't compile this file if the C++ version is less than C++17
#include "shared.h"
#if __cplusplus >= 201703L

#include <clickhouse/client.h>
#include <clickhouse/columns/string.h>
#include <clickhouse/query.h>

#include "tsdb.h"
// #include <fmt/core.h>
// #include <fmt/format.h>
// #include <spdlog/spdlog.h>

#include "internal_shared.hpp"

namespace tsdb_cpp {

using tsdb_shared::logger;

template <typename F, typename ErrRecoverFn>
inline typename std::decay<decltype(std::declval<F>()())>::type ex_log_boundary(F&& f, ErrRecoverFn&& err_recover_fn) try {
    return f();
} catch (const clickhouse::Error& e) {
    logger().error("Unhandled stck exception: {}", e.what());
    return err_recover_fn();
} catch (const std::exception& e) {
    logger().error("Unhandled generic exception: {}", e.what());
    return err_recover_fn();
} catch (...) {
    logger().error("Unhandled unknown exception");
    return err_recover_fn();
}

// Forward declaration
void tsdb_entry_notify_batch_close(tsdb_entry_impl* entry);
std::string_view tsdb_entry_get_correlation_id(tsdb_entry_impl* entry) noexcept;
clickhouse::Client& tsdb_entry_get_client(tsdb_entry_impl* entry);

struct tsdb_batch_insertion::impl {
    impl(tsdb_entry_impl* parent, std::string_view table_name, std::string_view metric_name)
        : parent_(parent),
          table_name_(table_name),
          metric_name_(metric_name),
          col_metric_name_(std::make_shared<clickhouse::ColumnString>()),
          col_point_name_(std::make_shared<clickhouse::ColumnString>()),
          col_value_(std::make_shared<clickhouse::ColumnFloat64>()),
          col_nanoseconds_(std::make_shared<clickhouse::ColumnInt64>()) {
        correlation_id_ = tsdb_shared::random_uuid();
        logger().debug("TSDB-lf batch created from entry {}, correlation ID: {}", tsdb_entry_get_correlation_id(parent_), correlation_id_);
    }
    ~impl() {
        logger().debug("TSDB-lf batch closed, correlation ID: {}", correlation_id_);
        // Notify the parent that the batch is closed
        tsdb_entry_notify_batch_close(parent_);
    }

    bool add_point(std::string_view name, double value, long long nanoseconds) {
        // logger().trace("TSDB-lf batch point added, correlation ID: {}", correlation_id_);
        logger().trace("TSDB-lf batch point added, cur count = {}, correlation ID: {}", col_metric_name_->Size() + 1, correlation_id_);
        col_metric_name_->Append(metric_name_);
        col_point_name_->Append(name);
        col_value_->Append(value);
        col_nanoseconds_->Append(nanoseconds);
        return true;
    }

    bool commit() {
        logger().debug("TSDB-lf batch committing, correlation ID: {}", correlation_id_);
        clickhouse::Block block;
        block.AppendColumn("metric_name", col_metric_name_);
        block.AppendColumn("point_name", col_point_name_);
        block.AppendColumn("value", col_value_);
        block.AppendColumn("nanoseconds", col_nanoseconds_);
        tsdb_entry_get_client(parent_).Insert(table_name_, block);
        logger().debug("TSDB-lf batch committed, correlation ID: {}", correlation_id_);
        return true;
    }

    void err_log_env() {
        logger().error("TSDB-lf batch environment: table_name: `{}`, metric_name: `{}`, correlation ID: {}, data count: {}", table_name_,
                       metric_name_, correlation_id_, col_value_->Size());
    }

private:
    tsdb_entry_impl* const parent_;
    std::string table_name_;
    std::string metric_name_;
    std::shared_ptr<clickhouse::ColumnString> col_metric_name_;
    std::shared_ptr<clickhouse::ColumnString> col_point_name_;
    std::shared_ptr<clickhouse::ColumnFloat64> col_value_;
    std::shared_ptr<clickhouse::ColumnInt64> col_nanoseconds_;
    std::string correlation_id_;
};

tsdb_batch_insertion::tsdb_batch_insertion(std::nullptr_t) : impl_(nullptr) {
}
tsdb_batch_insertion::tsdb_batch_insertion(std::unique_ptr<impl> impl) : impl_(std::move(impl)) {
}
tsdb_batch_insertion::~tsdb_batch_insertion() {
    close();
}
tsdb_batch_insertion::tsdb_batch_insertion(tsdb_batch_insertion&& other) : impl_(std::move(other.impl_)) {
}
bool tsdb_batch_insertion::close() {
    if (auto impl = std::move(impl_)) {
        return ex_log_boundary([&] { return impl->commit(); },
                               [&] {
                                   impl_->err_log_env();
                                   return false;
                               });
    }
    return true;
}
bool tsdb_batch_insertion::add_point(std::string_view name, double value, long long nanoseconds) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->add_point(name, value, nanoseconds); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_batch_insertion::ensure_open() const {
    if (!impl_) {
        logger().error("TSDB-lf batch is not open");
        return false;
    }
    return true;
}

struct tsdb_entry_impl {
    tsdb_entry_impl(std::string_view remote, uint16_t port) : client_(clickhouse::ClientOptions().SetHost(remote.data()).SetPort(port)) {
        // Initialize the TSDB environment
        tsdb_shared::init_tsdb_env();

        correlation_id_ = tsdb_shared::random_uuid();
        logger().debug("TSDB-lf entry created, correlation ID: {}", correlation_id_);
        logger().debug("TSDB-lf using default table name `{}`", table_name_);
    }
    ~tsdb_entry_impl() { logger().debug("TSDB-lf entry closed, correlation ID: {}", correlation_id_); }

    bool close() {
        if (open_batch_count_ > 0) {
            // // Fail-fast
            // throw std::runtime_error("Cannot close tsdb_entry while there are open batches");
            logger().error("Cannot close tsdb_entry while there are open batches, correlation ID: {}", correlation_id_);
            logger().error("!!! This is a BUG in the client code, please fix it !!!");
            err_log_env();
            return false;
        }
        return true;
    }

    bool create_table_inner(std::string_view table_name, std::string_view engine, bool can_order, std::string_view extras = "") {
        auto purify_identifier = [](std::string_view ident) {
            auto space_pos = ident.find(' ');
            return space_pos == std::string_view::npos ? ident : ident.substr(0, space_pos);
        };

        set_active_table(purify_identifier(table_name));
        logger().debug("TSDB-lf creating table `{}`, correlation ID: {}", table_name, correlation_id_);
        auto query = fmt::format(
            "CREATE TABLE IF NOT EXISTS {} ("
            "metric_name LowCardinality(String) COMMENT 'Metric name',"
            "point_name LowCardinality(String) COMMENT 'Point name',"
            "tags Map(LowCardinality(String), LowCardinality(String)) COMMENT 'Point tags',"
            "value Float64 COMMENT 'Point value' Codec(Gorilla, LZ4),"
            "nanoseconds Int64 COMMENT 'Point time in nanoseconds' Codec(DoubleDelta, LZ4)"
            ") ENGINE = {}",
            table_name, engine);
        if (can_order) {
            query += " ORDER BY (metric_name, point_name, nanoseconds)";
        }
        if (!extras.empty()) {
            query += " ";
            query += extras;
        }
        client_.Execute(query);
        return true;
    }

    bool drop_table_inner(std::string_view table_name) {
        logger().debug("TSDB-lf dropping table `{}`, correlation ID: {}", table_name, correlation_id_);
        auto query = fmt::format("DROP TABLE IF EXISTS {}", table_name);
        client_.Execute(query);
        return true;
    }

    bool set_async_insert_buffer_size(size_t size) {
        logger().debug("TSDB-lf setting async insert buffer size to {}, correlation ID: {}", size, correlation_id_);
        async_insert_buffer_size_ = size;
        return true;
    }

    size_t get_async_insert_buffer_size() const { return async_insert_buffer_size_; }

    bool set_active_table(std::string_view table_name) {
        logger().debug("TSDB-lf setting active table to `{}`, correlation ID: {}", table_name, correlation_id_);
        table_name_ = table_name;
        return true;
    }

    bool set_metric_name(std::string_view metric_name) {
        logger().debug("TSDB-lf setting metric name to `{}`, correlation ID: {}", metric_name, correlation_id_);
        metric_name_ = metric_name;
        return true;
    }

    bool insert_point(const point& p) { return insert_point(p.name_, p.value_, p.nanoseconds_); }
    bool insert_point(std::string_view name, double value, long long nanoseconds) {
        logger().debug("TSDB-lf point inserting{}, correlation ID: {}", async_insert_buffer_size_ ? " (async)" : "", correlation_id_);
        if (async_insert_buffer_size_ == 0) {
            clickhouse::Block block;
            auto col1 = std::make_shared<clickhouse::ColumnString>();
            col1->Append(metric_name_);
            block.AppendColumn("metric_name", col1);
            auto col2 = std::make_shared<clickhouse::ColumnString>();
            col2->Append(name);
            block.AppendColumn("point_name", col2);
            auto col3 = std::make_shared<clickhouse::ColumnFloat64>();
            col3->Append(value);
            block.AppendColumn("value", col3);
            auto col4 = std::make_shared<clickhouse::ColumnInt64>();
            col4->Append(nanoseconds);
            block.AppendColumn("nanoseconds", col4);
            client_.Insert(table_name_, block);
            logger().debug("TSDB-lf point inserted, correlation ID: {}", correlation_id_);
            return true;
        }
        auto escape_sql_literal = [](std::string_view str) {
            std::string result;
            result.reserve(str.size() + 2);
            result.push_back('\'');
            for (char c : str) {
                if (c == '\'' || c == '\\') {
                    result.push_back('\\');
                }
                result.push_back(c);
            }
            result.push_back('\'');
            return result;
        };
        auto sql = fmt::format("INSERT INTO {} (metric_name, point_name, value, nanoseconds) VALUES ({}, {}, {}, {})", table_name_,
                               escape_sql_literal(metric_name_), escape_sql_literal(name), value, nanoseconds);
        clickhouse::Query query(sql);
        query.SetSetting("async_insert", clickhouse::QuerySettingsField{"1", 1});
        if (++async_insert_cur_count_ >= async_insert_buffer_size_) {
            async_insert_cur_count_ = 0;
            query.SetSetting("wait_for_async_insert", clickhouse::QuerySettingsField{"1", 1});
        } else {
            query.SetSetting("wait_for_async_insert", clickhouse::QuerySettingsField{"0", 1});
        }
        query.SetSetting("async_insert_busy_timeout_ms", clickhouse::QuerySettingsField{"5000", 1});
        query.SetSetting("async_insert_max_data_size", clickhouse::QuerySettingsField{"104857600", 1});
        // query.SetSetting("async_insert_use_adaptive_busy_timeout", clickhouse::QuerySettingsField{"0", 1});
        client_.Execute(query);
        logger().debug("TSDB-lf point inserted, correlation ID: {}", correlation_id_);
        return true;
    }

    bool insert_points(const std::vector<point>& points) {
        logger().debug("TSDB-lf [!LEGACY!] points inserting, correlation ID: {}", correlation_id_);
        auto batch = new_batch_insert();
        for (const auto& p : points) {
            batch.add_point(p.name_, p.value_, p.nanoseconds_);
        }
        batch.close();
        logger().debug("TSDB-lf [!LEGACY!] points inserted, correlation ID: {}", correlation_id_);
        return true;
    }

    tsdb_batch_insertion new_batch_insert() {
        auto batch = tsdb_batch_insertion(std::make_unique<tsdb_batch_insertion::impl>(this, table_name_, metric_name_));
        if (!batch) {
            logger().error("Failed to create TSDB-lf batch, correlation ID: {}", correlation_id_);
            return nullptr;
        }
        open_batch_count_++;
        return batch;
    }

    friend void tsdb_entry_notify_batch_close(tsdb_entry_impl* entry) {
        if (entry->open_batch_count_ == 0) {
            // Fail-fast
            throw std::runtime_error("Cannot close tsdb_entry while there are open batches");
        }
        entry->open_batch_count_--;
    }
    friend std::string_view tsdb_entry_get_correlation_id(tsdb_entry_impl* entry) noexcept { return entry->correlation_id_; }
    friend clickhouse::Client& tsdb_entry_get_client(tsdb_entry_impl* entry) { return entry->client_; }

    void err_log_env() {
        logger().error("TSDB-lf entry environment: table_name: `{}`, metric_name: `{}`, correlation ID: {}, open_batch_count: {}",
                       table_name_, metric_name_, correlation_id_, open_batch_count_);
    }

private:
    clickhouse::Client client_;
    std::string table_name_          = "tsdb_cpp_dist";
    std::string metric_name_         = "";
    size_t async_insert_buffer_size_ = 0;
    size_t async_insert_cur_count_   = 0;
    size_t open_batch_count_         = 0;
    std::string correlation_id_;
};

tsdb_entry::tsdb_entry(std::nullptr_t) : impl_(nullptr) {
}
tsdb_entry::tsdb_entry(std::string_view remote, uint16_t port) : impl_{nullptr} {
    // Initialize the TSDB environment
    tsdb_shared::init_tsdb_env();

    logger().debug("TSDB-lf entry connecting to stck at {}:{}", remote, port);

    ex_log_boundary(
        [&] {
            impl_ = std::make_unique<tsdb_entry_impl>(remote, port);
            return true;
        },
        [&] {
            if (impl_) {
                impl_->err_log_env();
            }
            return false;
        });
}
tsdb_entry::~tsdb_entry() {
    close();
}
tsdb_entry::tsdb_entry(tsdb_entry&& other) : impl_(std::move(other.impl_)) {
}
bool tsdb_entry::close() {
    if (auto impl = std::move(impl_)) {
        return ex_log_boundary([&] { return impl->close(); },
                               [&] {
                                   impl->err_log_env();
                                   return false;
                               });
    }
    return true;
}
bool tsdb_entry::create_table() {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->create_table_inner("tsdb_cpp", "MergeTree()", true); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::create_memory_table() {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->create_table_inner("tsdb_cpp_memory", "Memory()", false); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::drop_table() {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->drop_table_inner("tsdb_cpp"); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::drop_memory_table() {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->drop_table_inner("tsdb_cpp_memory"); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::create_dist_table(std::string_view cluster_name) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary(
        [&] {
            std::string table_name = "tsdb_cpp";
            if (!cluster_name.empty()) {
                table_name += fmt::format(" ON CLUSTER {}", cluster_name);
            }
            if (!impl_->create_table_inner(table_name, "ReplicatedMergeTree('/clickhouse/tables/{shard}/tsdb_cpp', '{replica}')", true)) {
                logger().error("Failed to create table `{}`", table_name);
                impl_->err_log_env();
                return false;
            }
            table_name = "tsdb_cpp_dist";
            if (!cluster_name.empty()) {
                table_name += fmt::format(" ON CLUSTER {}", cluster_name);
            }
            auto engine = fmt::format("Distributed({}, default, tsdb_cpp, rand())", cluster_name);
            if (!impl_->create_table_inner(table_name, engine, false)) {
                logger().error("Failed to create table `{}`", table_name);
                impl_->err_log_env();
                return false;
            }
            return true;
        },
        [&] {
            impl_->err_log_env();
            return false;
        });
}
bool tsdb_entry::set_async_insert_buffer_size(size_t size) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->set_async_insert_buffer_size(size); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
size_t tsdb_entry::get_async_insert_buffer_size() const {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->get_async_insert_buffer_size(); }, [] { return 0; });
}
bool tsdb_entry::set_active_table(std::string_view table_name) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->set_active_table(table_name); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::set_metric_name(std::string_view metric_name) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->set_metric_name(metric_name); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::insert_point(const point& p) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->insert_point(p); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::insert_points(const std::vector<point>& points) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->insert_points(points); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::insert_point(std::string_view name, double value, long long nanoseconds) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->insert_point(name, value, nanoseconds); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
tsdb_batch_insertion tsdb_entry::new_batch_insert() {
    if (!ensure_open()) {
        return nullptr;
    }
    return ex_log_boundary([&] { return impl_->new_batch_insert(); },
                           [&] {
                               impl_->err_log_env();
                               return nullptr;
                           });
}
bool tsdb_entry::ensure_open() const {
    if (!impl_) {
        logger().error("TSDB-lf entry is not open");
        return false;
    }
    return true;
}

}  // namespace tsdb_cpp

#endif  // __cplusplus >= 201703L
