#include "tsdb_hf.h"

#include <fmt/chrono.h>
#include <fmt/format.h>
#include <fmt/std.h>
#include <sys/stat.h>
#include <zstd.h>

#include <cstddef>
#include <cstdint>
#include <cstring>
#include <exception>
#include <nlohmann/json.hpp>
#include <nlohmann/json_fwd.hpp>
#include <string_view>

#include "internal_shared.hpp"
#include "shared.h"

using namespace tsdb_preludes;

/// @brief Custom C++ wrapper for ZSTD.
namespace zstd_cpp {

struct zstd_error : std::runtime_error {
    zstd_error(const std::string& msg) : std::runtime_error(msg) {}
    zstd_error(const char* msg) : std::runtime_error(msg) {}
};

void check_zstd(size_t code, string_view msg = "ZSTD error") {
    if (ZSTD_isError(code)) {
        throw zstd_error(fmt::format("{}: {}", msg, ZSTD_getErrorName(code)));
    }
}

struct CCtx {
    CCtx() : cctx_(ZSTD_createCCtx()) {
        if (!cctx_) {
            throw zstd_error("Failed to create ZSTD_CCtx");
        }
    }
    CCtx(std::nullptr_t) : cctx_(nullptr) {}
    ~CCtx() { ZSTD_freeCCtx(cctx_); }
    CCtx(const CCtx&) = delete;
    CCtx(CCtx&& other) : CCtx(nullptr) { std::swap(cctx_, other.cctx_); }
    CCtx& operator=(CCtx other) {
        std::swap(cctx_, other.cctx_);
        return *this;
    }

    void reset() { ZSTD_CCtx_reset(cctx_, ZSTD_reset_session_only); }

    void set_compression_level(int compressionLevel) {
        check_zstd(ZSTD_CCtx_setParameter(cctx_, ZSTD_c_compressionLevel, compressionLevel));
    }

    /// @brief Performs a streaming compression operation.
    /// @param output
    /// @param input
    /// @param endOp
    /// @return The approximate remaining bytes to flush to the output buffer.
    size_t compress_stream(ZSTD_outBuffer& output, ZSTD_inBuffer& input, ZSTD_EndDirective endOp) {
        size_t remaining;
        remaining = ZSTD_compressStream2(cctx_, &output, &input, endOp);
        check_zstd(remaining);
        return remaining;
    }

    /// @brief Performs a simple streaming compression operation. This is a convenience function
    /// to simplify handling of compressed data.
    /// @param src The source data. Specify `nullptr` to end the current compression session.
    /// @param srcSize The size of the source data.
    /// @param dst The destination buffer. Only used as a temporary buffer.
    /// @param dstCapacity The capacity of the destination buffer.
    /// @param callback The callback function that will be called with the compressed data.
    /// Should have the signature `fn(const void* compressed_data, size_t size) -> void`.
    template <typename F>
    void compress_stream_simple(const void* src, size_t srcSize, void* dst, size_t dstCapacity, F&& callback) {
        ZSTD_inBuffer input   = {src, srcSize, 0};
        ZSTD_outBuffer output = {dst, dstCapacity, 0};

        if (src == nullptr) {
            input.size = 0;
            size_t remaining;
            // Must flush everything inside the internal buffer now.
            do {
                output.pos = 0;
                remaining  = compress_stream(output, input, ZSTD_e_end);
                callback(output.dst, output.pos);
            } while (remaining > 0);
            return;
        }

        // It's okay to stop as soon as we've drained the input buffer, and leave some
        // remaining compressed data in the ZSTD's internal buffer.
        while (input.pos < input.size) {
            output.pos = 0;
            compress_stream(output, input, ZSTD_e_continue);
            callback(output.dst, output.pos);
        }
    }

private:
    ZSTD_CCtx* cctx_;
};

}  // namespace zstd_cpp

namespace tsdb_hf_cpp {

// WARN: Linux-specific because of separator `/`
static std::string concat_path(string_view base, string_view path) {
    if (!path.empty() && path.front() == '/') {
        // Override the base path
        return std::string(path);
    }
    std::string result;
    result.reserve(base.size() + path.size() + 1);
    result.append(base);
    if (!base.empty() && base.back() != '/') {
        result.push_back('/');
    }
    result.append(path);
    return result;
}

// WARN: Linux-specific because of separator `/`
static std::string replace_extension(string_view path, string_view extension) {
    std::string result;
    // First strip the extension
    auto pos = path.find_last_of('.');
    if (pos == std::string::npos || path.substr(pos).find('/') != std::string::npos) {
        // No extension found
        result = path;
    } else {
        // Remove the extension (including the dot)
        result = path.substr(0, pos);
    }
    // Then append the new extension
    if (!extension.empty()) {
        if (extension.front() != '.') {
            result.push_back('.');
        }
        result.append(extension);
    }
    return result;
}

static bool is_file_exists(const std::string& path) {
    struct stat buffer;
    return stat(path.c_str(), &buffer) == 0;
}

static int do_mkdir(const char* path, mode_t mode) {
    struct stat st;
    int status = 0;

    if (stat(path, &st) != 0) {
        /* Directory does not exist. EEXIST for race condition */
        if (mkdir(path, mode) != 0 && errno != EEXIST) {
            status = -1;
        }
    } else if (!S_ISDIR(st.st_mode)) {
        errno  = ENOTDIR;
        status = -1;
    }

    return status;
}

/**
** mkpath - ensure all directories in path exist
** Algorithm takes the pessimistic view and works top-down to ensure
** each directory in path exists, rather than optimistically creating
** the last element and working backwards.
*/
static int mkpath(const char* path, mode_t mode) {
    char* pp;
    char* sp;
    int status;
    char* copypath = strdup(path);
    if (copypath == nullptr) {
        return -1;
    }

    status = 0;
    pp     = copypath;
    while (status == 0 && (sp = strchr(pp, '/')) != 0) {
        if (sp != pp) {
            /* Neither root nor double slash in path */
            *sp    = '\0';
            status = do_mkdir(copypath, mode);
            *sp    = '/';
        }
        pp = sp + 1;
    }
    if (status == 0) {
        status = do_mkdir(path, mode);
    }
    free(copypath);
    return status;
}

/// @brief Renames a file to a new extension. If the extension is empty,
/// the file will be renamed to have no extension.
static void rename_to_extension(const std::string& path, string_view extension = "") {
    auto new_path = replace_extension(path, extension);
    if (rename(path.c_str(), new_path.c_str()) != 0) {
        throw_errno("Failed to rename file");
    }
}

// Forward declaration
void tsdb_entry_notify_stream_close(tsdb_entry::impl* entry);
string_view tsdb_entry_get_correlation_id(tsdb_entry::impl* entry) noexcept;
size_t tsdb_entry_get_part_size(tsdb_entry::impl* entry) noexcept;

struct StreamMetadata {
    std::string metricName;
    std::string name;
    uint64_t timestampOffset;
    uint64_t interval;
};

struct tsdb_entry_stream::impl {
    impl(tsdb_entry::impl* parent, std::string stream_base_dir, StreamMetadata metadata)
        : parent_(parent),
          stream_base_dir_(std::move(stream_base_dir)),
          metadata_(std::move(metadata)),
          part_max_size_(tsdb_entry_get_part_size(parent_)) {
        correlation_id_ = tsdb_shared::random_uuid();
        // Claim the stream lock
        lockfile_ = bin_file(concat_path(stream_base_dir_, ".lockfile"), "wb");
        lockfile_.lock();
        logger().debug("TSDB-hf stream created from entry {}, correlation ID: {}", tsdb_entry_get_correlation_id(parent_), correlation_id_);

        // TODO: Maybe recycle cctx_?
        cctx_.reset();

        rewrite_metadata();
    }
    ~impl() {
        logger().debug("TSDB-hf stream closed, correlation ID: {}", correlation_id_);

        // Notify the parent that the stream is closed
        tsdb_entry_notify_stream_close(parent_);
    }

    // WARN: You must call this method exactly once before dropping the object.
    bool close() {
        // Flush the remaining data
        close_current_part();

        // Emit the metadata about the stream and the data parts
        rewrite_metadata();

        return true;
    }

    bool insert_point(double value) { return insert_points(&value, 1); }
    bool insert_points(const double* values, size_t count) {
        const auto count_copy = count;
        logger().debug("TSDB-hf stream inserting {} points, correlation ID: {}", count_copy, correlation_id_);
        while (count > 0) {
            if (cur_part_remaining_size_ == 0) {
                open_new_part();
            }

            auto to_write = std::min(count, cur_part_remaining_size_);
            logger().trace("TSDB-hf stream inserting {} points for this pass, correlation ID: {}", to_write, correlation_id_);
            cctx_.compress_stream_simple(values, to_write * sizeof(double), buffer_, sizeof buffer_,
                                         [this](const void* data, size_t size) { write_to_disk(data, size); });
            cur_part_remaining_size_ -= to_write;
            cur_points_count_ += to_write;
            values += to_write;
            count -= to_write;
        }
        logger().debug("TSDB-hf stream inserted {} points, correlation ID: {}", count_copy, correlation_id_);

        return true;
    }

    void err_log_env() {
        logger().error(
            "TSDB-hf stream environment: data_dir: `{}`, part_size: {}, metric_name: `{}`, correlation ID: {}, stream_base_dir: `{}`, "
            "written points: {}, parts_count: {}",
            stream_base_dir_, part_max_size_, metadata_.metricName, correlation_id_, stream_base_dir_, cur_points_count_, cur_parts_index_);
    }

private:
    nlohmann::json dump_metadata() {
        return {
            {"metric_name", metadata_.metricName}, {"name", metadata_.name},          {"timestamp_offset", metadata_.timestampOffset},
            {"interval", metadata_.interval},      {"parts_count", cur_parts_index_}, {"points_per_part", part_max_size_},
            {"total_points", cur_points_count_},
        };
    }

    void rewrite_metadata() {
        logger().debug("TSDB-hf stream rewriting metadata, correlation ID: {}", correlation_id_);
        auto metadata      = dump_metadata().dump(4);
        auto metadata_path = concat_path(stream_base_dir_, "metadata.json.part");
        auto metadata_file = bin_file(metadata_path, "wb");
        metadata_file.write(metadata.data(), metadata.size());
        metadata_file.close();
        rename_to_extension(metadata_path);
        logger().debug("TSDB-hf stream rewritten metadata, correlation ID: {}", correlation_id_);
    }

    void write_to_disk(const void* data, size_t size) {
        cur_part_file_.write(data, size);

        // TODO: Maybe flush eagerly?
        // if (cur_part_file_) {
        //     cur_part_file_.flush();
        // }
        // if (cur_part_remaining_size == 0) {
        //     close_current_part();
        // }
    }

    void open_new_part() {
        close_current_part();

        auto path = concat_path(stream_base_dir_, fmt::format("part_{:09d}.zst.part", cur_parts_index_));
        logger().trace("TSDB-hf stream opening part file `{}`, correlation ID: {}", path, correlation_id_);
        if (is_file_exists(path)) {
            throw std::runtime_error(fmt::format("Part file already exists: {}", path));
        }
        cur_part_file_           = bin_file(path, "wb");
        cur_part_file_path_      = std::move(path);
        cur_part_remaining_size_ = part_max_size_;
        cur_parts_index_++;
    }

    void close_current_part() {
        if (cur_part_file_) {
            // Flush the remaining data
            cctx_.compress_stream_simple(nullptr, 0, buffer_, sizeof(buffer_),
                                         [this](const void* data, size_t size) { write_to_disk(data, size); });

            // Commit the previous part (remove the `.part` extension)
            logger().trace("TSDB-hf stream closing part file, correlation ID: {}", correlation_id_);
            cur_part_file_.close();
            rename_to_extension(cur_part_file_path_);
        }
    }

    tsdb_entry::impl* const parent_;
    StreamMetadata metadata_;
    zstd_cpp::CCtx cctx_;
    std::string stream_base_dir_;
    std::string cur_part_file_path_;
    bin_file cur_part_file_;
    bin_file lockfile_;
    // NOTE: The size represents the remaining points count, not the remaining bytes.
    size_t cur_parts_index_         = 0;
    size_t cur_part_remaining_size_ = 0;
    const size_t part_max_size_     = -1;
    size_t cur_points_count_        = 0;
    char buffer_[16 * 1024];
    std::string correlation_id_;
};

tsdb_entry_stream::tsdb_entry_stream(std::nullptr_t) : impl_(nullptr) {
}
tsdb_entry_stream::tsdb_entry_stream(std::unique_ptr<impl> impl) : impl_(std::move(impl)) {
}
tsdb_entry_stream::~tsdb_entry_stream() {
    try {
        close();
    } catch (const std::exception& e) {
        fprintf(stderr, "Unhandled exception in ~tsdb_entry_stream: %s\n", e.what());
    }
}
tsdb_entry_stream::tsdb_entry_stream(tsdb_entry_stream&& other) : impl_(std::move(other.impl_)) {
}
bool tsdb_entry_stream::close() {
    if (auto impl = std::move(impl_)) {
        return ex_log_boundary([&] { return impl->close(); },
                               [&] {
                                   impl->err_log_env();
                                   return false;
                               });
    }
    return true;
}
bool tsdb_entry_stream::insert_point(double value) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->insert_point(value); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry_stream::insert_points(const double* values, size_t count) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->insert_points(values, count); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry_stream::ensure_open() const {
    if (!impl_) {
        logger().error("TSDB-hf stream is not open");
        return false;
    }
    return true;
}

struct tsdb_entry::impl {
    impl() {
        // Initialize the TSDB environment
        tsdb_shared::init_tsdb_env();

        correlation_id_ = tsdb_shared::random_uuid();
        logger().debug("TSDB-hf entry created, correlation ID: {}", correlation_id_);
    }
    ~impl() { logger().debug("TSDB-hf entry closed, correlation ID: {}", correlation_id_); }

    // WARN: You must call this method exactly once before dropping the object.
    bool close() {
        if (open_streams_count_ > 0) {
            // // Fail-fast
            // throw std::runtime_error("Cannot close tsdb_entry while there are open streams");
            logger().error("Cannot close tsdb_entry while there are open streams, correlation ID: {}", correlation_id_);
            logger().error("!!! This is a BUG in the client code, please fix it !!!");
            err_log_env();
            return false;
        }
        return true;
    }

    bool set_data_dir(string_view data_dir) {
        // TODO: Maybe sanitize the path?
        logger().debug("TSDB-hf entry setting data dir to `{}`, correlation ID: {}", data_dir, correlation_id_);
        data_dir_ = data_dir;
        return true;
    }

    bool set_part_size(size_t size) {
        logger().debug("TSDB-hf entry setting part size to {}, correlation ID: {}", size, correlation_id_);
        if (size < 1) {
            throw std::invalid_argument("Part size must be positive");
        }
        part_size_ = size;
        return true;
    }

    bool set_metric_name(string_view metric_name) {
        logger().debug("TSDB-hf entry setting metric name to `{}`, correlation ID: {}", metric_name, correlation_id_);
        metric_name_ = metric_name;
        return true;
    }

    tsdb_entry_stream new_stream(std::string name, uint64_t timestampOffset, uint64_t interval) {
        // Obtain a lock on the stream directory to prevent watch-daemon from reading incomplete folders
        bin_file stream_lockfile = bin_file(concat_path(data_dir_, ".streams.lockfile"), "wb");
        stream_lockfile.lock();

        auto dir_name        = fmt::format("{}-{:%Y-%m-%d_%H-%M-%S}", name, std::chrono::system_clock::now());
        auto stream_base_dir = concat_path(data_dir_, dir_name);
        auto stream_metadata = StreamMetadata{metric_name_, std::move(name), timestampOffset, interval};
        // Ensure base dir
        check_errno(mkpath(stream_base_dir.c_str(), 0777) >= 0, "Failed to create stream base dir");
        auto stream =
            tsdb_entry_stream(tsdb_shared::make_unique<tsdb_entry_stream::impl>(this, stream_base_dir, std::move(stream_metadata)));
        open_streams_count_++;
        return stream;
    }

    friend void tsdb_entry_notify_stream_close(tsdb_entry::impl* entry) {
        if (entry->open_streams_count_ == 0) {
            // Fail-fast
            throw std::runtime_error("open_streams_count_ is already 0");
        }
        entry->open_streams_count_--;
    }
    friend string_view tsdb_entry_get_correlation_id(tsdb_entry::impl* entry) noexcept { return entry->correlation_id_; }
    friend size_t tsdb_entry_get_part_size(tsdb_entry::impl* entry) noexcept { return entry->part_size_; }

    void err_log_env() {
        logger().error(
            "TSDB-hf entry environment: data_dir: `{}`, part_size: {}, metric_name: `{}`, correlation ID: {}, open_streams_count: {}",
            data_dir_, part_size_, metric_name_, correlation_id_, open_streams_count_);
    }

private:
    std::string data_dir_ = "data";
    // 1'000'000
    size_t part_size_          = 1000000;
    size_t open_streams_count_ = 0;
    std::string metric_name_;
    std::string correlation_id_;
};

tsdb_entry::tsdb_entry(std::nullptr_t) : impl_(nullptr) {
}
tsdb_entry::tsdb_entry() : impl_(tsdb_shared::make_unique<impl>()) {
}
tsdb_entry::~tsdb_entry() {
    close();
}
tsdb_entry::tsdb_entry(tsdb_entry&& other) : impl_(std::move(other.impl_)) {
}
bool tsdb_entry::close() {
    if (auto impl = std::move(impl_)) {
        return ex_log_boundary([&] { return impl->close(); },
                               [&] {
                                   impl->err_log_env();
                                   return false;
                               });
    }
    return true;
}
bool tsdb_entry::set_data_dir(string_view data_dir) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->set_data_dir(data_dir); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::set_part_size(size_t size) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->set_part_size(size); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
bool tsdb_entry::set_metric_name(string_view metric_name) {
    if (!ensure_open()) {
        return false;
    }
    return ex_log_boundary([&] { return impl_->set_metric_name(metric_name); },
                           [&] {
                               impl_->err_log_env();
                               return false;
                           });
}
tsdb_entry_stream tsdb_entry::new_stream(std::string name, uint64_t timestampOffset, uint64_t interval) {
    if (!ensure_open()) {
        return nullptr;
    }
    return ex_log_boundary([&] { return impl_->new_stream(name, timestampOffset, interval); },
                           [&] {
                               impl_->err_log_env();
                               return nullptr;
                           });
}
bool tsdb_entry::ensure_open() const {
    if (!impl_) {
        logger().error("TSDB-hf entry is not open");
        return false;
    }
    return true;
}

}  // namespace tsdb_hf_cpp
